import glob
import logging
import os.path
import time
from abc import ABC, abstractmethod
from typing import Any, Iterable, List, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from tqdm import tqdm
from tqdm.autonotebook import trange

logger = logging.getLogger(__name__)


class BaseRetriever(ABC):
    """Base class for retrieval."""

    @abstractmethod
    def search(self, query: str, top_k: int, batch_size: int = -1) -> str:
        """search the str, return the top list maybe with score"""

    def ingest(self, document):
        return

    def similarity_search_by_vector(self, query_embedding: List[float], k: int = 10, **kwargs: Any):
        """Perform ANN search by vector."""
        pass

    def similarity_search_by_text(self, text: str, text_embedder, k: int = 10, **kwargs: Any):
        """Perform ANN search by text."""
        pass


class AutoModelForRetrieval(object):
    def __init__(
        self,
        embedding_model: Optional[nn.Module] = None,
        reranker_model: Optional[nn.Module] = None,
        method: Literal['cosine', 'knn', 'llm'] = "cosine",
    ) -> None:
        super().__init__()
        self.embedding_model = embedding_model
        self.reranker_model = reranker_model
        self.method = method

    def search(
        self,
        query_embed: Union[torch.Tensor, np.ndarray],
        document_embed: Optional[torch.Tensor] = None,
        index_path: Optional[str] = None,
        top_k: int = 3,
        batch_size: int = -1,
        convert_to_numpy: bool = True,
        **kwargs,
    ) -> Optional[Tuple[np.ndarray, np.ndarray]]:
        """
        index_path: it can be an index file, or a folder with more index files
        """
        if len(query_embed.shape) == 1:
            if isinstance(query_embed, np.ndarray):
                query_embed = query_embed.reshape(1, -1)
            else:
                query_embed = query_embed.view(1, -1)

        self.top_k = top_k
        if document_embed is None and index_path is None:
            logging.warning('Please provide either document_embed for knn/tensor search or index_path for faiss search')
            return

        if index_path is not None:
            import faiss

            start_time = time.time()
            if not isinstance(index_path, (list, tuple)) and os.path.isfile(index_path):
                faiss_index = faiss.read_index(index_path)
                logger.info(f'Loading faiss index successfully, elapsed time: {time.time()-start_time:.2}s')
                faiss_retrieval = FaissRetrieval(faiss_index)
            else:
                if isinstance(index_path, (list, tuple)) and os.path.isfile(index_path[0]):
                    index_files = index_path
                else:
                    index_files = glob.glob(index_path)

                logger.info(
                    f'Loading index successfully, files: {len(index_files)}, elapsed: {time.time() - start_time:.2}s'
                )
                # TODO: not so good to generalize using only torch.load
                first_file = torch.load(index_files[0])
                faiss_retrieval = FaissRetrieval(first_file)
                for i in range(1, len(index_files)):
                    faiss_retrieval.add(index_files[i])

            if batch_size < 1:
                dists, indices = faiss_index.search(query_embed.astype(np.float32), k=top_k)
            else:
                dists, indices = faiss_retrieval.search(
                    query_embed=query_embed,
                    top_k=top_k,
                    batch_size=batch_size,
                )
            return dists, indices

        elif self.method == "knn":
            dists, indices = knn_search(query_embed, document_embed, top_k)

        elif self.method == "cosine":
            dists, indices = cosine_similarity_search(
                query_embed, document_embed, top_k=top_k, batch_size=batch_size, convert_to_numpy=convert_to_numpy
            )

        else:
            raise ValueError(
                f"Only 'cosine' and 'knn' method are supported by similarity_search, while get {self.method}"
            )

        return dists, indices

    def similarity(self, queries: Union[str, List[str]], keys: Union[str, List[str], np.ndarray]):
        return

    def get_relevant_documents(self, query: str):
        return

    def get_pandas_candidate(
        self,
        query_ids: Union[pd.Series, np.ndarray],
        document_ids: Union[pd.Series, np.ndarray],
        dists: np.ndarray,
        indices: np.ndarray,
    ) -> pd.DataFrame:
        if isinstance(query_ids, pd.Series):
            query_ids = query_ids.values
        if isinstance(document_ids, pd.Series):
            document_ids = document_ids.values

        retrieval = {
            'query_id': np.repeat(query_ids, self.top_k),
            'predict_id': document_ids[indices.ravel()],
            'score': dists.ravel(),
        }
        return pd.DataFrame(retrieval)

    def get_rerank_df(
        self,
        input_df: pd.DataFrame,
        query_key: str = 'query_id',
        document_key: str = 'document_id',
        predict_key: str = 'predict_id',
    ) -> pd.DataFrame:
        """
        1: the candidate is in ground truth pool
        2: the candidate is related or not with query
        """
        logger.info('Generate rerank samples based on the retrieval prediction with: query_id, document_ids, pred_ids')

        samples = []
        for _, row in tqdm(input_df.iterrows(), total=len(input_df)):
            query_id = row[query_key]
            document_ids = row[document_key]
            if pd.isna(row[predict_key]):
                print(f' Data error, no {predict_key} in input_df')
                continue
            else:
                predict_ids_list = row[predict_key].split()

            if pd.isna(document_ids):
                print(f' Data Error, the ground truth {document_key} is none')
                for id in predict_ids_list:
                    samples.append({query_key: query_id, document_key: id, 'label': 0})
            else:
                documents = document_ids.split()
                for id in predict_ids_list:
                    if id not in documents:
                        samples.append({query_key: query_id, document_key: id, 'label': 0})
                    else:
                        samples.append({query_key: query_id, document_key: id, 'label': 1})

        return pd.DataFrame(samples)


def knn_search(query_embed, document_embed, top_k):
    from sklearn.neighbors import NearestNeighbors

    neighbors_model = NearestNeighbors(n_neighbors=top_k, metric="cosine", n_jobs=-1)
    neighbors_model.fit(document_embed)
    dists, indices = neighbors_model.kneighbors(query_embed)
    return dists, indices


def cosine_similarity_search(
    query_embed: torch.Tensor,
    document_embed: torch.Tensor,
    top_k: int = 1,
    batch_size: int = 128,
    penalty: bool = True,
    temperature: float = 0,
    convert_to_numpy: bool = True,
    show_progress_bar: bool = None,
):
    if len(query_embed.size()) == 1:
        query_embed = query_embed.view(1, -1)
    assert query_embed.size()[1] == document_embed.size()[1], (
        f"The embed Shape of query_embed and document_embed should be same, "
        f"while received query {query_embed.size()} and document {document_embed.size()}"
    )
    chunk = batch_size if batch_size > 0 else len(query_embed)
    embeddings_chunks = query_embed.split(chunk)

    dists = []
    indices = []
    for idx in trange(0, len(embeddings_chunks), desc="Batches", disable=not show_progress_bar):
        cos_sim_chunk = torch.matmul(embeddings_chunks[idx], document_embed.transpose(0, 1))
        cos_sim_chunk = torch.nan_to_num(cos_sim_chunk, nan=0.0)
        # if penalty:
        #     pen = ((contents["old_source_count"].values==0) & (contents["old_nonsource_count"].values==1))
        #     cos_sim_chunk = cos_sim_chunk.clamp(0,1) ** torch.Tensor((1 + pen*0.5).reshape(1,-1)).to(DEVICE)
        #     cos_sim_chunk = cos_sim_chunk.clamp(0,1) ** torch.Tensor((1-(contents["old_count"].values==0)*0.1).
        #     reshape(1,-1)).to(DEVICE)
        if temperature:
            cos_sim_chunk = cos_sim_chunk / temperature
        top_k = min(top_k, cos_sim_chunk.size(1))
        dists_chunk, indices_chunk = torch.topk(cos_sim_chunk, k=top_k, dim=1)
        dists.append(dists_chunk[:, :].detach().cpu())
        indices.append(indices_chunk[:, :].detach().cpu())
    dists = torch.cat(dists)
    indices = torch.cat(indices)
    if convert_to_numpy:
        dists = dists.numpy()
        indices = indices.numpy()
    return dists, indices


class FaissRetrieval(BaseRetriever):
    def __init__(self, corpus_index: Union[np.ndarray, torch.Tensor]):
        import faiss

        if isinstance(corpus_index, (np.ndarray, torch.Tensor)):
            index = faiss.IndexFlatIP(corpus_index.shape[1])
            self.index = index
        else:
            self.index = corpus_index

    def add(self, corpus_embed, corpus_ids=None):
        """incremental add embed"""
        corpus_embed = np.array(corpus_embed, dtype=np.float32)
        if corpus_embed.ndim == 1:
            corpus_embed = np.expand_dims(corpus_embed, axis=0)

        if corpus_ids:
            self.index.add_with_ids(corpus_embed, np.array(corpus_ids))
        else:
            self.index.add(corpus_embed)

    def search(
        self,
        query_embeddings: Union[torch.Tensor, np.ndarray],
        top_k: int = 100,
        batch_size: int = 128,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        1. Encode queries into dense embeddings
        2. Search through faiss index

        To get document_id: np.array([[int(document_lookup[idx]) for idx in indices] for indices in all_indices])
        """
        query_size = len(query_embeddings)
        assert query_size > 0, 'Please make sure the query_embeddings is not empty'

        all_scores = []
        all_indices = []

        for i in tqdm(range(0, query_size, batch_size), desc="Searching"):
            j = min(i + batch_size, query_size)
            query_embed = query_embeddings[i:j]
            score, index = self.index.search(query_embed.astype(np.float32), k=top_k)
            all_scores.append(score)
            all_indices.append(index)

        all_scores = np.concatenate(all_scores, axis=0)
        all_indices = np.concatenate(all_indices, axis=0)

        return all_scores, all_indices

    def combine(self, results: Iterable[Tuple[np.ndarray, np.ndarray]]):
        import faiss

        rh = None
        for scores, indices in results:
            if rh is None:
                print(f'Initializing Heap. Assuming {scores.shape[0]} queries.')
                rh = faiss.ResultHeap(scores.shape[0], scores.shape[1])
            rh.add_result(-scores, indices)
        rh.finalize()
        corpus_scores, corpus_indices = -rh.D, rh.I

        return corpus_scores, corpus_indices


class BM25Retrieval(BaseRetriever):
    def __init__(
        self, documents, chunk_size=None, chunk_overlap=None, splitter=None, stop_words_dir: Optional[str] = None
    ):
        from rank_bm25 import BM25Okapi

        self.documents = documents
        self.splitter = splitter
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.stop_words_dir = stop_words_dir

        if self.splitter:
            documents = self.splitter.split_text(self.documents)
        else:
            documents = self.documents

        self.bm25 = BM25Okapi(documents)

    def search(self, query: str, top_k: int, batch_size: int = -1) -> List[Tuple[List[str], float]]:
        scores = self.bm25.get_scores(query)
        sorted_docs = sorted(zip(self.documents, scores), key=lambda x: x[1], reverse=True)[:top_k]
        return sorted_docs

    def _load_stop_words(self):
        stop_words_path = os.path.join(self.stop_words_dir, 'stop_words.txt')
        if not os.path.exists(stop_words_path):
            raise Exception(f"system stop words: {stop_words_path} not found")

        stop_words = []
        with open(stop_words_path, 'r', encoding='utf8') as reader:
            for line in reader:
                line = line.strip()
                stop_words.append(line)
        return stop_words


class ElasticRetriever(BaseRetriever):
    def __init__(self):
        super(ElasticRetriever, self).__init__()
        from elasticsearch import Elasticsearch

        self.es = Elasticsearch()


class EnsembleRetriever(BaseRetriever):
    """
    RRF_fusion
    """

    def __init__(self, retrievers: List[BaseRetriever], weights=None):
        self.retrievers = retrievers

    def search(self, query: str, top_k: int = 10, batch_size: int = -1) -> List[str]:
        combined_results = []
        for retriever in self.retrievers:
            combined_results.extend(retriever.search(query, top_k))

        unique_results = list(set(combined_results))
        unique_results.sort()

        return unique_results[:top_k]


class GraphRetrieval(BaseRetriever):
    def __init__(self, index_name):
        super(GraphRetrieval, self).__init__()

    def search(self, query: str, top_k: int, batch_size: int = -1) -> str:
        pass

    def global_search(self):
        pass

    def local_search(self):
        pass



"""
`SimCSE: Simple Contrastive Learning of Sentence Embeddings
<https://arxiv.org/abs/2104.08821>`_
"""

import logging
from typing import Callable, Optional, Union

import torch
import torch.nn.functional as F
from torch import nn

logger = logging.getLogger(__name__)


class SimCSE(nn.Module):
    """Cosine similarity = normalize + inner product"""

    def __init__(
        self,
        criterion: Union[nn.Module, Callable] = nn.CrossEntropyLoss(label_smoothing=0.0, reduction='mean'),
        temperature: float = 0.05,
        dynamic_temperature: bool = False,
        **kwargs
    ):
        super().__init__()
        self.criterion = criterion
        self.temperature = temperature
        if dynamic_temperature:
            # TODO
            self.temperature = nn.Parameter(torch.tensor(temperature))

    def forward(
        self,
        query_embeddings: torch.Tensor,
        pos_embeddings: torch.Tensor,
        neg_embeddings: Optional[torch.Tensor] = None,
    ):
        similarity = F.cosine_similarity(query_embeddings.unsqueeze(1), pos_embeddings.unsqueeze(0), dim=-1)

        if neg_embeddings is not None:
            neg_similarity = F.cosine_similarity(query_embeddings.unsqueeze(1), neg_embeddings.unsqueeze(0), dim=-1)
            similarity = torch.cat([similarity, neg_similarity], dim=1)

        similarity = similarity / self.temperature

        target = torch.arange(0, query_embeddings.size(0), dtype=torch.long, device=query_embeddings.device)

        loss = self.criterion(similarity, target)
        return loss



from typing import Dict, List


def get_map(qid2positive: Dict[str, List[str]], qid2ranking: Dict[str, List[str]], cutoff_rank: int = 10):
    """
    qid2positive: {qid: [pos1_doc_id, pos2_doc_id]}
    qid2ranking: {qid: [rank1_doc_id, rank2_doc_id, rank3_doc_id]}
    """

    def average_precision(positives: List[str], ranked_doc_ids: List[str], cutoff: int) -> float:
        """
        for each cut_off, calculate its precision
        """
        hits = 0
        sum_precisions = 0.0

        for rank, doc_id in enumerate(ranked_doc_ids[:cutoff], start=1):
            if doc_id in positives:
                hits += 1
                sum_precisions += hits / rank

        return sum_precisions / len(positives) if positives else 0.0

    qid2map = dict()

    for qid in qid2positive:
        positives = qid2positive[qid]
        ranked_doc_ids = qid2ranking[qid]
        qid2map[qid] = average_precision(positives, ranked_doc_ids, cutoff_rank)

    return {f"map@{cutoff_rank}": sum(qid2map.values()) / len(qid2ranking.keys())}



from typing import Dict, List, Tuple, Union


def get_mrr(qid2positive: Dict[str, List[str]], qid2ranking: Dict[str, List[str]], cutoff_rank: int = 10):
    """
    qid2positive: {qid: [pos1_doc_id, pos2_doc_id]}
    qid2ranking: {qid: [rank1_doc_id, rank2_doc_id, rank3_doc_id]}
    """
    qid2mrr = dict()

    for qid in qid2positive:
        positives = qid2positive[qid]
        ranked_doc_ids = qid2ranking[qid]

        for rank, doc_id in enumerate(ranked_doc_ids, start=1):
            if doc_id in positives:
                if rank <= cutoff_rank:
                    qid2mrr[qid] = 1.0 / rank
                break
    return {f"mrr@{cutoff_rank}": sum(qid2mrr.values()) / len(qid2ranking.keys())}



from typing import Callable, Optional, Union

import torch
import torch.nn as nn


class PairKL(nn.Module):
    def __init__(
        self,
        criterion: Union[nn.Module, Callable, None] = nn.KLDivLoss(reduction="batchmean"),
        temperature: float = 0.05,
        **kwargs
    ):
        super().__init__()
        self.criterion = criterion
        self.temperature = temperature

    def forward(
        self, query_embeddings: torch.Tensor, positive_embeddings: torch.Tensor, scores: torch.Tensor, **kwargs
    ):
        similarity = torch.einsum('bn,bn->b', query_embeddings, positive_embeddings)
        similarity = similarity / self.temperature
        similarity = torch.log_softmax(similarity, dim=-1)
        target = torch.softmax(scores / self.temperature, dim=-1)

        return self.criterion(similarity, target)



import logging
from typing import List, Optional

import numpy as np
import torch
from torch.utils.data import Dataset, DistributedSampler
from torch.utils.data.sampler import BatchSampler, Sampler

logger = logging.getLogger(__name__)


class GroupedBatchSampler(BatchSampler):
    """
    It enforces that the batch only contain elements from the same group.
    shuffle should be set to False for custom sampler
    """

    def __init__(
        self, sampler: Sampler, group_ids: List[int], batch_size: int, shuffle: bool = True, seed: Optional[int] = None
    ):
        """
        sampler (Sampler): Base sampler.
        group_ids (list[int]): If the sampler produces indices in range [0, N),
            `group_ids` must be a list of `N` ints which contains the group id of each sample.
            The group ids must be a set of integers in the range [0, num_groups).
        batch_size (int): Size of mini-batch.
        """
        self.sampler = sampler
        self.group_ids = np.asarray(group_ids)
        assert self.group_ids.ndim == 1
        self.batch_size = batch_size
        self.shuffle = shuffle
        groups = np.unique(self.group_ids).tolist()

        # buffer the indices of each group until batch size is reached
        self.buffer_per_group = {k: [] for k in groups}
        self.rng = np.random.default_rng(seed)

    def __iter__(self):
        for idx in self.sampler:
            group_id = self.group_ids[idx]
            group_buffer = self.buffer_per_group[group_id]
            group_buffer.append(idx)
            if len(group_buffer) == self.batch_size:
                yield group_buffer[:]  # yield a copy of the list
                del group_buffer[:]

    def __len__(self):
        return (len(self.sampler) + self.batch_size - 1) // self.batch_size

    def _create_batches(self):
        indices = [idx for idx in self.sampler]
        self.rng.shuffle(indices)

        for idx in indices:
            group_id = self.group_ids[idx]

            group_buffer = self.buffer_per_group[group_id]
            visited = self.visited_per_group[group_id]

            added = False
            for i, visited_set in enumerate(visited):
                added = True
                group_buffer[i].append(idx)
                if len(group_buffer[i]) == self.batch_size:
                    self.batches.append(group_buffer[i])
                    group_buffer[i] = None
                    visited[i] = None
                break

            self.buffer_per_group[group_id] = [x for x in group_buffer if x is not None]
            self.visited_per_group[group_id] = [x for x in visited if x is not None]
            if not added:
                self.buffer_per_group[group_id].append([idx])
                self.visited_per_group[group_id].append(set())

        if self.shuffle:
            self.rng.shuffle(self.batches)

    def _clear(self):
        self.batches = []
        self.buffer_per_group = {k: [] for k in self.groups}
        self.visited_per_group = {k: [] for k in self.groups}


class GroupSortedBatchSampler(BatchSampler):
    def __init__(
        self, sampler, group_ids, batch_size, shuffle: bool = True, drop_last: bool = False, seed: Optional[int] = None
    ):
        self.sampler = sampler
        self.group_ids = np.asarray(group_ids)
        assert self.group_ids.ndim == 1
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.drop_last = drop_last
        groups = np.unique(self.group_ids).tolist()

        self.batches = []
        # buffer the indices of each group until batch size is reached
        self.buffer_per_group = {k: [] for k in groups}
        self.rng = np.random.default_rng(seed)

    def __iter__(self):
        if not self.batches:
            self._create_batches()
        for batch in self.batches:
            yield batch
        self._clear()

    def __len__(self):
        if not self.batches:
            self._create_batches()
        return len(self.batches)

    def _create_batches(self):
        indices = []
        group_id_map = dict()
        group_idx = 0
        group_indices = []

        for idx in self.sampler:
            indices.append(idx)

        # if self.shuffle:
        #     self.rng.shuffle(indices)

        for idx in indices:
            group_id = self.group_ids[idx]
            if group_id not in group_id_map:
                group_id_map[group_id] = group_idx
                group_idx += 1
            group_indices.append(group_id_map[group_id])

        l = list(zip(group_indices, indices))
        l.sort()

        l = np.asarray([i[-1] for i in l])
        batches = split_batches(l, self.batch_size, drop_last=self.drop_last)
        for batch in batches:
            self.batches.append(list(batch))

    def _clear(self):
        self.batches = []
        # self.buffer_per_group = {k: [] for k in self.groups}


class DistributedBucketSampler(torch.utils.data.distributed.DistributedSampler):
    def __init__(
        self,
        dataset,
        batch_size,
        boundaries,
        num_replicas=None,
        rank=None,
        shuffle=True,
        expand_new_speaker=False,
        new_speaker_ratio=None,
        speaker_uniform_batch=False,
        drop_last=False,
    ):
        pass

    def _create_buckets(self):
        pass

    def __iter__(self):
        pass

    def __len__(self):
        return


def split_batches(inputs, batch_size: int, drop_last: bool = False):
    l = np.split(inputs, np.arange(batch_size, len(inputs), batch_size))
    if drop_last:
        if len(l[-1]) != batch_size:
            l = l[:-1]
    return l


class SyncedSampler(DistributedSampler):
    def __init__(
        self,
        dataset: Dataset,
        num_replicas: Optional[int] = None,
        rank: Optional[int] = None,
        shuffle: bool = True,
        seed: int = 0,
    ):
        super(SyncedSampler, self).__init__(dataset, num_replicas, rank, shuffle, seed)
        self.num_samples = len(self.dataset)
        self.total_size = len(self.dataset)

    def __iter__(self):
        if self.shuffle:
            # deterministically shuffle based on epoch and seed
            g = torch.Generator()
            g.manual_seed(self.seed + self.epoch)
            indices = torch.randperm(len(self.dataset), generator=g).tolist()
        else:
            indices = list(range(len(self.dataset)))

        # DO NOT SUB SAMPLE!
        assert len(indices) == self.total_size
        assert len(indices) == self.num_samples

        return iter(indices)

    def set_epoch(self, epoch: int):
        super(SyncedSampler, self).set_epoch(epoch)
        logger.info(f'Setting Data Sampler Epoch to {epoch}')



from setuptools import Command, find_packages, setup

version_file = 'src/retrievals/version.py'


def get_version():
    with open(version_file, 'r') as f:
        exec(compile(f.read(), version_file, 'exec'))
    return locals()['__version__']


base_packages = [
    "torch>=1.6.0",
    "transformers >= 4.0.0",
    "tokenizers>=0.14",
    "datasets>=1.1.3",
    "tqdm >= 4.66",
]


eval = ["beir >= 2.0.0", "mteb", "C-MTEB>=1.1.0"]


dev = [
    "mkdocs-material == 9.2.8",
    "mkdocs-awesome-pages-plugin == 2.9.2",
    "mkdocs-jupyter == 0.24.7",
    "faiss-gpu",
    "peft",
    "accelerate>=0.20.0",
]


setup(
    name="open-retrievals",
    version=get_version(),
    author="Longxing Tan",
    author_email="tanlongxing888@163.com",
    description="Text Embeddings for Retrieval and RAG based on transformers",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    keywords="NLP retrieval RAG rerank text embedding contrastive",
    license="Apache 2.0 License",
    url="https://github.com/LongxingTan/open-retrievals",
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    package_data={"": ["**/*.cu", "**/*.cpp", "**/*.cuh", "**/*.h", "**/*.pyx"]},
    zip_safe=False,
    python_requires=">=3.7.0",
    install_requires=base_packages,
    extras_require={"eval": base_packages + eval, "dev": base_packages + eval + dev},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



"""File OCR"""

import logging
import time
from abc import ABC, abstractmethod

logger = logging.getLogger(__name__)


class OcrConfig:
    ocr_model = 'ppocr'


class OCRecognizer(object):
    """OCR class for recognize the scanned document"""

    def __init__(self, ocr_model='ppocr'):
        if ocr_model == 'ppocr':
            self.ocr_model = PPRecognizer()

    def recognize(self, file):
        logging.info('START OCR')
        start_time = time.time()

        elapsed_time = time.time() - start_time
        logging.info(f'FINISH OCR, Elapsed time: {elapsed_time:.3f}s')
        return

    def apply_preprocess(self, data, preprocessors):
        """preprocess set"""
        for preprocessor in preprocessors:
            data = preprocessor.process(data)
        return data

    def apply_postprocess(self, data, postprocessors):
        """post-process set"""
        for postprocessor in postprocessors:
            data = postprocessor.process(data)
        return data

    def _structure(self, img):
        """structure analysis"""
        return

    def _ocr(self, img):
        """ocr function"""
        return


class PreProcessor(ABC):
    """base preprocess"""

    def __init__(self) -> None:
        pass

    def process(self, example):
        raise NotImplementedError("Subclasses should implement this method")


class PostProcessor(ABC):
    """Base class for OCR post-processing"""

    def __init__(self) -> None:
        pass

    def process(self, example):
        raise NotImplementedError("Subclasses should implement this method")


class PPRecognizer:
    """A class to encapsulate PaddleOCR functionality for text recognition in images."""

    def __init__(self, use_angle_cls=False, lang="ch"):
        from paddleocr import PaddleOCR, PPStructure

        self.ocr_model = PaddleOCR(use_angle_cls=use_angle_cls, lang=lang)

    def recognize(self, img):
        try:
            result = self.ocr_model.ocr(img, cls=False)
            return result
        except Exception as e:
            print(f"An error occurred during OCR: {e}")
            return None



import torch
from torch import distributed as dist
from torch import nn


class DistributedLoss(nn.Module):
    def __init__(self, criterion):
        super(DistributedLoss, self).__init__()
        self.criterion = criterion
        self.word_size = dist.get_world_size()
        self.rank = dist.get_rank()

    def forward(self, x: torch.Tensor, y: torch.Tensor):
        dist_x = self.gather_tensor(x)
        dist_y = self.gather_tensor(y)
        loss = self.criterion(dist_x, dist_y)
        return loss

    def gather_tensor(self, t: torch.Tensor):
        gathered = [torch.empty_like(t) for _ in range(self.word_size)]
        dist.all_gather(gathered, t)
        gathered[self.rank] = t
        return torch.cat(gathered, dim=0)



"""Some default prompts"""

RAG_PROMPT = """
"""


SUMMARIZE_PROMPT = """
Generating a comprehensive summary of the data provided below.
Given one or two entities, and a list of descriptions, all related to the same entity or group of entities.
Please concatenate all of these into a single, comprehensive description.

#####
Entities: {entity_name}
Description List: {description_list}
#####
Output:
"""


QUERY_GENERATION_PROMPT = """\
Context information is below.

---------------------
{context_str}
---------------------

Given the context information and not prior knowledge.
generate only questions based on the below query.

You are a Teacher/ Professor. Your task is to setup \
{num_questions_per_chunk} questions for an upcoming \
quiz/examination. The questions should be diverse in nature \
across the document. Restrict the questions to the \
context information provided."
"""



"""Build hard negative for retrieval/reranking"""

import os
import random
from argparse import ArgumentParser

parser = ArgumentParser()
parser.add_argument("--input", type=str, required=True)
parser.add_argument("--output", type=str, required=True)

args = parser.parse_args()


def load_ranking(rank_file, relevance, num_hn_sample, depth):
    """
    rank_file: the output file from retrieve/rerank, each line with query_id, passage_id, score, rank
    relevance: hash table record the true passage_id for each query_id
    n_sample: numbers of hard negative
    depth: the retrieval/rerank chosen numbers from the rank_file result
    """
    with open(rank_file) as f:
        lines = iter(f)
        qid, pid, _, _ = next(lines).strip().split()

        curr_q = qid
        negatives = [] if pid in relevance[pid] else [pid]

        while True:
            try:
                q, p, _, _ = next(lines).strip().split()
                if q != curr_q:
                    # 开始下一个query_id之前, 首先把之前的返回；再重新从q开始
                    negatives = negatives[:depth]
                    random.shuffle(negatives)
                    yield curr_q, relevance[curr_q], negatives[:num_hn_sample]

                    curr_q = q
                    negatives = [] if p in relevance[q] else [p]
                else:
                    if p not in relevance[qid]:
                        negatives.append(p)

            except StopIteration:
                negatives = negatives[:depth]
                random.shuffle(negatives)
                yield curr_q, relevance[curr_q], negatives[:num_hn_sample]
                return


if __name__ == "__main__":
    os.makedirs(args.output, exist_ok=True)

    with open("hard_negative.json", "w") as f:
        for x in load_ranking():
            f.write(x + "\n")



from typing import Callable, Optional, Union

import torch
import torch.nn as nn


class ColbertLoss(nn.Module):
    def __init__(
        self,
        criterion: Union[nn.Module, Callable] = nn.CrossEntropyLoss(label_smoothing=0.0, reduction='mean'),
        temperature: float = 0.02,
        use_inbatch_negative: bool = False,
    ):
        super(ColbertLoss, self).__init__()
        self.criterion = criterion
        self.temperature = temperature
        self.use_inbatch_negative = use_inbatch_negative

    def forward(
        self,
        query_embeddings: torch.Tensor,
        positive_embeddings: torch.Tensor,
        negative_embeddings: Optional[torch.Tensor] = None,
        query_attention_mask: Optional[torch.Tensor] = None,
    ):

        if negative_embeddings is None and self.use_inbatch_negative is False:
            raise ValueError(
                "No negative samples for ColBERT, either provide negative_embeddings or use_inbatch_negative=True"
            )

        scores = self.similarity(query_embeddings, positive_embeddings)

        if negative_embeddings is not None:
            negative_scores = self.similarity(
                query_embeddings, negative_embeddings, query_attention_mask=query_attention_mask
            )
            scores = torch.cat([scores, negative_scores], dim=-1)

        if self.temperature is not None:
            scores = scores / self.temperature

        if self.use_inbatch_negative:
            labels = torch.arange(scores.size(0), dtype=torch.long, device=scores.device)
        else:
            labels = torch.zeros(scores.size(0), dtype=torch.long, device=scores.device)

        loss = self.criterion(scores, labels)
        return loss

    def similarity(self, query_embeddings, document_embeddings, query_attention_mask=None):
        if self.use_inbatch_negative:  # query_embeddings.size(0) != document_embeddings.size(0) and
            late_interactions = torch.einsum(
                "bsh,cdh->bcsd",
                query_embeddings,
                document_embeddings,
            )
        else:
            document_embeddings = document_embeddings.view(
                query_embeddings.size(0), -1, document_embeddings.size(1), document_embeddings.size(2)
            )

            late_interactions = torch.einsum(
                "bsh,bdth->bdst",
                query_embeddings,
                document_embeddings,
            )
        late_interactions = late_interactions.max(-1).values.sum(-1)

        # if query_attention_mask is not None:
        #     query_sequence_length = query_attention_mask[:, 1:].sum(-1, keepdim=False)
        #     if late_interactions.dim() == 2:  # if the train_group_size > 2, the late_interactions shape: batch * neg
        #         query_sequence_length = query_sequence_length.unsqueeze(1)
        #
        #     late_interactions = late_interactions / query_sequence_length
        # else:
        #     late_interactions = late_interactions / query_embeddings.size(1)
        return late_interactions



import logging
from typing import Dict

import torch
from torch import nn

logger = logging.getLogger(__name__)


class FGM:
    """Fast Gradient Sign Attack"""

    def __init__(self, model: nn.Module):
        self.model = model
        self.backup: Dict[str, torch.Tensor] = dict()

    def attack(self, epsilon: float = 1.0, emb_name: str = "word_embeddings"):
        """
        emb_name: model's embedding name
        """
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name:
                self.backup[name] = param.data.clone()
                norm = torch.norm(param.grad)
                if norm != 0:
                    r_at = epsilon * param.grad / norm
                    param.data.add_(r_at)

    def restore(self, emb_name: str = "word_embeddings"):
        """
        emb_name: model's embedding name
        """
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name:
                assert name in self.backup
                param.data = self.backup[name]
        self.backup.clear()


class EMA:
    """Exponential moving average

    Example:

        # init:
        ema = EMA(model, 0.999)
        ema.register()

        # train:
        def train():
            optimizer.step()
            ema.update()

        # eval:
        def evaluate():
            ema.apply_shadow()
            # evaluation code here
            ema.restore()
    """

    def __init__(self, model: nn.Module, decay: float = 0.999):
        self.model = model
        self.decay = decay
        self.shadow = dict()
        self.backup = dict()

    def register(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                self.shadow[name] = param.data.clone()

    def update(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                assert name in self.shadow
                new_average = (1.0 - self.decay) * param.data + self.decay * self.shadow[name]
                self.shadow[name] = new_average.clone()

    def apply_shadow(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                assert name in self.shadow
                self.backup[name] = param.data
                param.data = self.shadow[name]

    def restore(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                assert name in self.backup
                param.data = self.backup[name]
        self.backup = dict()


class PGD:
    def __init__(self, model: nn.Module):
        self.model = model
        self.emb_backup = dict()
        self.grad_backup = dict()

    def attack(
        self, epsilon: float = 1.0, alpha: float = 0.3, emb_name: str = "word_embeddings", is_first_attack: bool = False
    ):
        """
        emb_name: set to your model's embedding name
        """
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name:
                if is_first_attack:
                    self.emb_backup[name] = param.data.clone()
                norm = torch.norm(param.grad)
                if norm != 0:
                    r_at = alpha * param.grad / norm
                    param.data.add_(r_at)
                    param.data = self.project(name, param.data, epsilon)

    def restore(self, emb_name: str = "word_embeddings"):
        """
        emb_name: set to your model's embedding name
        """
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name:
                assert name in self.emb_backup
                param.data = self.emb_backup[name]
        self.emb_backup = dict()

    def project(self, param_name, param_data, epsilon):
        r = param_data - self.emb_backup[param_name]
        if torch.norm(r) > epsilon:
            r = epsilon * r / torch.norm(r)
        return self.emb_backup[param_name] + r

    def backup_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                self.grad_backup[name] = param.grad.clone()

    def restore_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                param.grad = self.grad_backup[name]


class AWP:
    def __init__(
        self,
        model: nn.Module,
        optimizer,
        adv_param: str = "weight",
        adv_lr: float = 0.0001,
        adv_eps: float = 0.001,
        adv_start_epoch: int = 0,
        adv_step: int = 1,
        scaler=None,
    ):
        self.model = model
        self.optimizer = optimizer
        self.adv_param = adv_param
        self.adv_lr = adv_lr
        self.adv_eps = adv_eps
        self.adv_start_epoch = adv_start_epoch
        self.adv_step = adv_step
        self.backup = dict()
        self.backup_eps = dict()
        self.scaler = scaler

    def attack_backward(self, inputs, criterion, labels, epoch: int):
        if self.adv_lr == 0 or epoch < self.adv_start_epoch:
            return

        logger.info('[AWP] start')
        self._save()
        for i in range(self.adv_step):
            self._attack_step()
            with torch.cuda.amp.autocast():
                y_preds = self.model(inputs)
                adv_loss = criterion(y_preds.view(-1, 1), labels.view(-1, 1))
                # adv_loss, logits = self.model(input_ids=x, attention_mask=attention_mask, labels=y)
                # adv_loss = adv_loss.mean()
            self.optimizer.zero_grad()
            self.scaler.scale(adv_loss).backward()

        self._restore()

    def _attack_step(self):
        e = 1e-6
        for name, param in self.model.named_parameters():
            if param.requires_grad and param.grad is not None and self.adv_param in name:
                norm1 = torch.norm(param.grad)
                norm2 = torch.norm(param.data.detach())
                if norm1 != 0 and not torch.isnan(norm1):
                    r_at = self.adv_lr * param.grad / (norm1 + e) * (norm2 + e)
                    param.data.add_(r_at)
                    param.data = torch.min(
                        torch.max(param.data, self.backup_eps[name][0]),
                        self.backup_eps[name][1],
                    )
                # param.data.clamp_(*self.backup_eps[name])

    def _save(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and param.grad is not None and self.adv_param in name:
                if name not in self.backup:
                    self.backup[name] = param.data.clone()
                    grad_eps = self.adv_eps * param.abs().detach()
                    self.backup_eps[name] = (
                        self.backup[name] - grad_eps,
                        self.backup[name] + grad_eps,
                    )

    def _restore(
        self,
    ):
        for name, param in self.model.named_parameters():
            if name in self.backup:
                param.data = self.backup[name]
        self.backup = dict()
        self.backup_eps = dict()



"""
https://arxiv.org/abs/2002.10857
"""

from typing import Optional

import torch
import torch.nn as nn


class MultiLabelCircleLoss(nn.Module):
    def __init__(self, reduction='mean', **kwargs):
        super().__init__()
        self.reduction = reduction
        self.inf = 1e12

    def forward(self, logits: torch.Tensor, labels: torch.Tensor, mask: Optional[torch.Tensor] = None):
        logits = (1 - 2 * labels) * logits
        logits_neg = logits - labels * self.inf
        logits_pos = logits - (1 - labels) * self.inf
        zeros = torch.zeros_like(logits[..., :1])
        logits_neg = torch.cat([logits_neg, zeros], dim=-1)
        logits_pos = torch.cat([logits_pos, zeros], dim=-1)
        neg_loss = torch.logsumexp(logits_neg, dim=-1)
        pos_loss = torch.logsumexp(logits_pos, dim=-1)
        loss = neg_loss + pos_loss
        if mask is not None:
            loss = loss / (mask.sum(-1).float() + 1e-12)
        if self.reduction == 'mean':
            loss = loss.mean()
        return loss



import json
import logging
import os
from copy import deepcopy
from typing import TYPE_CHECKING, Callable, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
from tqdm.auto import tqdm, trange
from transformers import (
    AutoModel,
    AutoModelForCausalLM,
    AutoModelForSequenceClassification,
    AutoTokenizer,
    PreTrainedTokenizer,
)
from transformers.modeling_outputs import SequenceClassifierOutput

from ..data.collator import LLMRerankCollator, RerankCollator
from ..losses.colbert_loss import ColbertLoss
from .base import Base
from .pooling import AutoPooling
from .utils import (
    batch_to_device,
    check_causal_lm,
    find_all_linear_names,
    get_device_name,
)

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)


class AutoModelForRanking(Base):
    def __init__(
        self,
        model: Optional[nn.Module] = None,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        train_group_size: int = 1,
        pooling_method: Optional[str] = None,
        loss_fn: Union[nn.Module, Callable] = None,
        loss_type: Literal['classification', 'regression'] = 'classification',
        max_length: Optional[int] = None,
        causal_lm: bool = False,
        task_prompt: Optional[str] = None,
        query_instruction: Optional[str] = None,
        document_instruction: Optional[str] = None,
        temperature: Optional[float] = None,
        device: Optional[str] = None,
        **kwargs,
    ):
        super().__init__()
        if isinstance(model, str):
            assert ValueError("Please use AutoModelForRanking.from_pretrained(model_name_or_path)")

        self.model: Optional[nn.Module] = model
        self.tokenizer = tokenizer
        self.train_group_size = train_group_size
        self.pooling_method = pooling_method
        if pooling_method:
            self.pooling = AutoPooling(self.pooling_method)

        self.loss_fn = loss_fn
        self.loss_type = loss_type
        self.causal_lm = causal_lm
        self.task_prompt = task_prompt
        self.query_instruction = query_instruction if query_instruction else ""
        self.document_instruction = document_instruction if document_instruction else ""

        if max_length is None:
            if (
                hasattr(self.model, "config")
                and hasattr(self.model.config, "max_position_embeddings")
                and hasattr(self.tokenizer, "model_max_length")
            ):
                max_length = min(self.model.config.max_position_embeddings, self.tokenizer.model_max_length)

        self.max_length = max_length
        self.temperature = temperature

        if device is None:
            self.device = get_device_name()
        else:
            self.device = device

        # self._post_init()
        self.to(self.device)

    # def _post_init(self):
    #     num_features: int = self.model.config.hidden_size
    #     self.classifier = nn.Linear(num_features, 1)
    #     try:
    #         state_dict = torch.load(os.path.join('./', "colbert_linear"), map_location=self.device)
    #         self.dense_pooler.load_state_dict(state_dict)
    #     except FileNotFoundError:
    #         self._init_weights(self.classifier)
    #         logger.warning("Could not find dense weight, initialize it randomly")

    def _init_weights(self, module: nn.Module):
        if isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=self.model.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.model.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def encode(
        self, input_ids: torch.Tensor, attention_mask: torch.Tensor
    ) -> Union[torch.Tensor, SequenceClassifierOutput]:
        # input_ids -> embedding
        model_output: SequenceClassifierOutput = self.model(input_ids, attention_mask, output_hidden_states=True)

        if not self.pooling_method:
            return model_output

        if 'last_hidden_state' in model_output:
            last_hidden_state = model_output['last_hidden_state']
        elif 'hidden_states' not in model_output:
            last_hidden_state = model_output[0]
        else:
            raise ValueError

        if self.pooling is not None:
            embeddings = self.pooling(last_hidden_state, attention_mask)
        else:
            embeddings = self.classifier(last_hidden_state)
        return embeddings

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        labels: Optional[torch.Tensor] = None,
        return_dict: Optional[bool] = True,
        **kwargs,
    ) -> Union[Dict[str, torch.Tensor], Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:
        if self.causal_lm:
            # LLM rerank
            model_output = self.model(input_ids, attention_mask, output_hidden_states=True)
            loss = self.loss_fn(model_output.logits, labels)
            outputs_dict = dict()
            outputs_dict['logits'] = model_output.logits
            outputs_dict['loss'] = loss
            return outputs_dict

        # cross-encoder rerank
        features = self.encode(input_ids=input_ids, attention_mask=attention_mask)

        if self.temperature is not None:
            features = features / self.temperature

        if not self.training:
            return features

        logits = features.logits
        scores = logits.view(-1, self.train_group_size)

        if return_dict:
            outputs_dict = dict()
            outputs_dict['logits'] = logits

        if labels is not None:
            if not self.loss_fn:
                if self.loss_type == 'regression':
                    logits = torch.sigmoid(logits)
                    self.loss_fn = nn.MSELoss()

                elif self.loss_type == 'classification':
                    self.loss_fn = nn.BCEWithLogitsLoss(reduction='mean')

            loss = self.loss_fn(scores.squeeze(), labels.squeeze().float())
            if return_dict:
                outputs_dict['loss'] = loss
                return outputs_dict
            else:
                return logits, loss
        else:
            if return_dict:
                return outputs_dict
            return logits

    def set_model_type(self, model_type: Literal['cross-encoder', 'colbert'], **kwargs):
        model_type = model_type.lower().replace('-', '').replace('_', '')
        logger.info(f'Set model type to: {model_type}')
        model_class = {'crossencoder': self, 'colbert': ColBERT, 'llm': LLMRanker}
        model_class = model_class.get(model_type)
        return model_class(
            model=self.model,
            tokenizer=self.tokenizer,
            loss_fn=self.loss_fn,
            **kwargs,
        )

    def preprocess_pair(
        self,
        batch_sentence_pair: List[List[str]],
        max_length: int,
        padding: Union[str, bool] = 'max_length',
    ):
        if isinstance(batch_sentence_pair[0][0], str):
            batch = self.tokenizer(
                batch_sentence_pair,
                padding=True,
                truncation=True,
                max_length=max_length,
                return_tensors="pt",
            )
        else:
            batch = self.tokenizer.pad(
                batch_sentence_pair,
                padding=True,
                max_length=None,
                pad_to_multiple_of=None,
                return_tensors='pt',
            )
        batch_on_device = {k: v.to(self.device) for k, v in batch.items()}
        return batch_on_device

    @torch.no_grad()
    def compute_score(
        self,
        sentence_pairs: Union[List[Tuple[str, str]], Tuple[str, str]],
        batch_size: int = 16,
        max_length: int = 512,
        normalize: bool = False,
        show_progress_bar: bool = None,
        **kwargs,
    ):
        """
        preprocess -> score -> output
        """
        self.model.eval()
        if isinstance(sentence_pairs[0], str):
            sentence_pairs = [sentence_pairs]

        length_sorted_idx = np.argsort([-self._text_length(q) - self._text_length(p) for q, p in sentence_pairs])
        sentences_sorted = [sentence_pairs[idx] for idx in length_sorted_idx]

        all_scores: List[float] = []
        for batch_start in tqdm(
            range(0, len(sentences_sorted), batch_size), desc='Scoring', disable=not show_progress_bar
        ):
            batch_sentences = sentences_sorted[batch_start : batch_start + batch_size]
            batch_on_device = self.preprocess_pair(batch_sentences, max_length=max_length)
            scores = self.model(**batch_on_device, return_dict=True).logits.view(-1).float()

            if normalize:
                scores = torch.sigmoid(scores)
            all_scores.extend(scores.cpu().float().tolist())

        all_scores = [all_scores[idx] for idx in np.argsort(length_sorted_idx)]

        if len(all_scores) == 1:
            return all_scores[0]

        return all_scores

    @torch.no_grad()
    def rerank(
        self,
        query: str,
        documents: List[str],
        data_collator: Optional[RerankCollator] = None,
        batch_size: int = 16,
        chunk_max_length: int = 256,
        chunk_overlap: int = 48,
        max_chunks_per_doc: int = 100,
        normalize: bool = False,
        show_progress_bar: bool = None,
        return_dict: bool = True,
        **kwargs,
    ):
        if query is None or len(query) == 0 or len(documents) == 0:
            return {'rerank_documents': [], 'rerank_scores': []}

        splitter = DocumentSplitter(
            chunk_size=chunk_max_length, chunk_overlap=chunk_overlap, max_chunks_per_doc=max_chunks_per_doc
        )
        text_pairs, sentence_pairs_pids = splitter.create_documents(
            query,
            documents,
            tokenizer=self.tokenizer,
        )

        tot_scores = self.compute_score(
            sentence_pairs=text_pairs,
            data_collator=data_collator,
            batch_size=batch_size,
            normalize=normalize,
            show_progress_bar=show_progress_bar,
        )

        merge_scores = [0 for _ in range(len(documents))]
        for pid, score in zip(sentence_pairs_pids, tot_scores):
            merge_scores[pid] = max(merge_scores[pid], score)

        merge_scores_argsort = np.argsort(merge_scores)[::-1]
        sorted_document = []
        sorted_scores = []
        for mid in merge_scores_argsort:
            sorted_scores.append(merge_scores[mid])
            sorted_document.append(documents[mid])

        if return_dict:
            return {
                'rerank_document': sorted_document,
                'rerank_scores': sorted_scores,
                'rerank_ids': merge_scores_argsort.tolist(),
            }
        else:
            return sorted_document

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_path: str,
        pooling_method: Optional[str] = None,
        num_labels: int = 1,
        loss_fn: Union[nn.Module, Callable] = None,
        loss_type: Literal['classification', 'regression'] = 'classification',
        causal_lm: bool = False,
        trust_remote_code: bool = True,
        use_fp16: bool = False,
        use_lora: bool = False,
        lora_config=None,
        lora_path: Optional[str] = None,
        quantization_config=None,
        task_prompt: Optional[str] = None,
        query_instruction: Optional[str] = None,
        document_instruction: Optional[str] = None,
        device: Optional[str] = None,
        temperature: Optional[float] = None,
        **kwargs,
    ):
        tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path, return_tensors=False, trust_remote_code=trust_remote_code
        )

        if causal_lm or check_causal_lm(model_name_or_path):
            logger.info('Set model to AutoModelForCausalLM')
            model = AutoModelForCausalLM.from_pretrained(
                model_name_or_path, quantization_config=quantization_config, trust_remote_code=trust_remote_code
            )
        else:
            logger.info('Set model to  AutoModelForSequenceClassification')
            model = AutoModelForSequenceClassification.from_pretrained(
                model_name_or_path, num_labels=num_labels, trust_remote_code=trust_remote_code, **kwargs
            )

        if device is None:
            device = get_device_name()

        if use_fp16 and device != 'cpu':
            logger.info('Set model to fp16, please note that if you want fp16 during training, set training_args fp16')
            model.half()

        if use_lora:
            logger.info('Set fine-tuning to LoRA')
            from peft import LoraConfig, TaskType, get_peft_model

            if lora_config is None:
                lora_alpha = 64
                lora_dropout = 0.05
                target_modules = find_all_linear_names(model)
                lora_config = LoraConfig(
                    lora_alpha=lora_alpha,
                    lora_dropout=lora_dropout,
                    target_modules=target_modules,
                    bias='none',
                    task_type=TaskType.CAUSAL_LM,
                )

            model = get_peft_model(model, lora_config)
            model.print_trainable_parameters()

        if lora_path is not None:
            logger.info('Load pretrained with LoRA adapter')
            from peft import LoraConfig, PeftModel

            model = PeftModel.from_pretrained(model, lora_path)
            model = model.merge_and_unload()

        reranker = cls(
            model=model,
            tokenizer=tokenizer,
            pooling_method=pooling_method,
            device=device,
            loss_fn=loss_fn,
            loss_type=loss_type,
            temperature=temperature,
            causal_lm=causal_lm,
            task_prompt=task_prompt,
            query_instruction=query_instruction,
            document_instruction=document_instruction,
        )
        return reranker


class ColBERT(Base):
    def __init__(
        self,
        model: Optional[nn.Module] = None,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        linear_layer: Optional[nn.Module] = None,
        max_length: Optional[int] = None,
        loss_fn: Union[nn.Module, Callable] = None,
        temperature: float = 1.0,
        device: Optional[str] = None,
        **kwargs,
    ):
        super().__init__()
        self.model = model
        self.tokenizer = tokenizer
        self.linear = linear_layer

        self.loss_fn = loss_fn
        self.temperature = temperature
        self.max_length = max_length
        self.device = device or get_device_name()
        self.to(self.device)

    def forward(
        self,
        query_input_ids: Optional[torch.Tensor],
        query_attention_mask: Optional[torch.Tensor],
        pos_input_ids: Optional[torch.Tensor],
        pos_attention_mask: Optional[torch.Tensor],
        neg_input_ids: Optional[torch.Tensor] = None,
        neg_attention_mask: Optional[torch.Tensor] = None,
        return_dict: Optional[bool] = True,
        **kwargs,
    ):
        query_embedding = self._encode(query_input_ids, attention_mask=query_attention_mask, normalize_embeddings=True)
        positive_embedding = self._encode(pos_input_ids, attention_mask=pos_attention_mask, normalize_embeddings=True)

        if self.training:
            if not self.loss_fn:
                self.loss_fn = ColbertLoss(temperature=self.temperature, use_inbatch_negative=False)

            negative_embedding = None
            if neg_input_ids is not None:
                negative_embedding = self._encode(
                    neg_input_ids, attention_mask=neg_attention_mask, normalize_embeddings=True
                )

            loss = self.loss_fn(
                query_embedding, positive_embedding, negative_embedding, query_attention_mask=query_attention_mask
            )

            if return_dict:
                outputs_dict = dict()
                outputs_dict['loss'] = loss
                return outputs_dict
            return loss

        scores = self.score(query_embedding, positive_embedding, query_attention_mask=query_attention_mask)
        return scores

    def preprocess_pair(
        self,
        batch_sentence_pair: List[List[str]],
        query_max_length: int,
        document_max_length: int,
        padding='max_length',
    ):
        query_list = [item[0] for item in batch_sentence_pair]
        document_list = [item[1] for item in batch_sentence_pair]

        query_batch_tokens = self.tokenizer(
            query_list, padding=padding, truncation=True, max_length=query_max_length, return_tensors='pt'
        )
        query_batch_tokens_on_device = {k: v.to(self.device) for k, v in query_batch_tokens.items()}
        document_batch_tokens = self.tokenizer(
            document_list, padding=padding, truncation=True, max_length=document_max_length, return_tensors='pt'
        )
        document_batch_tokens_on_device = {k: v.to(self.device) for k, v in document_batch_tokens.items()}

        return {
            "query_input_ids": query_batch_tokens_on_device['input_ids'],
            "query_attention_mask": query_batch_tokens_on_device['attention_mask'],
            "doc_input_ids": document_batch_tokens_on_device['input_ids'],
            "doc_attention_mask": document_batch_tokens_on_device['attention_mask'],
        }

    def _encode(
        self, input_ids: torch.Tensor, attention_mask: torch.Tensor, normalize_embeddings: bool = True
    ) -> torch.Tensor:
        """encode the ids and mask to embedding"""
        outputs: SequenceClassifierOutput = self.model(
            input_ids,
            attention_mask=attention_mask,
            output_hidden_states=True,
            return_dict=True,
        )
        if hasattr(outputs, 'last_hidden_state'):
            # hidden_state shape: [batch, sequence_length, attention_dim]
            hidden_state = outputs.last_hidden_state
        else:
            hidden_state = outputs.hidden_states[1]

        # without the first the [CLS] token
        embeddings = self.linear(hidden_state[:, 1:])
        embeddings = embeddings * attention_mask[:, 1:].unsqueeze(-1).float()

        if normalize_embeddings:
            embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=-1)
        return embeddings

    def encode(
        self,
        sentences: Union[str, List[str], Tuple[str], 'pd.Series', np.ndarray],
        batch_size: int = 16,
        show_progress_bar: bool = None,
        output_value: str = "sentence_embedding",
        convert_to_numpy: bool = True,
        convert_to_tensor: bool = False,
        device: str = None,
        normalize_embeddings: bool = True,
    ):
        device = device or self.device
        self.model.eval()
        self.model.to(device)

        if show_progress_bar is None:
            show_progress_bar = (
                logger.getEffectiveLevel() == logging.INFO or logger.getEffectiveLevel() == logging.DEBUG
            )

        if convert_to_tensor:
            convert_to_numpy = False

        if output_value != "sentence_embedding":
            convert_to_tensor = False
            convert_to_numpy = False

        input_was_string = False
        if isinstance(sentences, str) or not hasattr(sentences, "__len__"):
            sentences = [sentences]
            input_was_string = True

        all_embeddings = []
        length_sorted_idx = np.argsort([-self._text_length(sentence) for sentence in sentences])
        sentences_sorted = [sentences[idx] for idx in length_sorted_idx]

        for start_index in trange(0, len(sentences), batch_size, desc="Batches", disable=not show_progress_bar):
            sentences_batch = sentences_sorted[start_index : start_index + batch_size]
            features = self.tokenizer(
                sentences_batch,
                padding=True,
                truncation=True,
                return_tensors="pt",
                max_length=self.max_length,
            )
            features = batch_to_device(features, device)

            with torch.no_grad():
                embeddings = self._encode(**features, normalize_embeddings=normalize_embeddings)
                embeddings = embeddings.detach()

                if convert_to_numpy:
                    embeddings = embeddings.cpu()

                all_embeddings.extend(embeddings)

        all_embeddings = [all_embeddings[idx] for idx in np.argsort(length_sorted_idx)]

        if convert_to_tensor:
            if len(all_embeddings) > 0:
                all_embeddings = torch.stack(all_embeddings)
            else:
                all_embeddings = torch.Tensor()
        elif convert_to_numpy:
            all_embeddings = np.asarray([emb.numpy() for emb in all_embeddings])

        if input_was_string:
            all_embeddings = all_embeddings[0]

        return all_embeddings

    @torch.no_grad()
    def compute_score(
        self,
        sentence_pairs: Union[List[Tuple[str, str]], Tuple[str, str]],
        batch_size: int = 16,
        max_length: int = 256,
        normalize: bool = False,
        show_progress_bar: bool = None,
        **kwargs,
    ):
        self.model.eval()
        if isinstance(sentence_pairs[0], str) and len(sentence_pairs) == 2:
            sentence_pairs = [sentence_pairs]

        scores_list: List[float] = []
        for i in tqdm(range(0, len(sentence_pairs), batch_size), desc="Scoring", disable=not show_progress_bar):
            batch_on_device = self.preprocess_pair(
                sentence_pairs[i : i + batch_size], query_max_length=max_length, document_max_length=max_length
            )
            query_embedding = self._encode(
                batch_on_device['query_input_ids'], batch_on_device['query_attention_mask'], normalize_embeddings=True
            )
            doc_embedding = self._encode(
                batch_on_device['doc_input_ids'], batch_on_device['doc_attention_mask'], normalize_embeddings=True
            )
            scores = self.score(
                query_embedding, doc_embedding, query_attention_mask=batch_on_device['query_attention_mask']
            )
            if normalize:
                scores = torch.sigmoid(scores)
            scores_list.extend(scores.cpu().numpy().tolist())

        if len(scores_list) == 1:
            return scores_list[0]
        return scores_list

    def score(
        self,
        query_embeddings: torch.Tensor,
        document_embeddings: torch.Tensor,
        query_attention_mask: Optional[torch.Tensor] = None,
    ):
        # pair embedding -> score, mask the pad token while calculate similarity score
        late_interactions = torch.einsum(
            "bsh,bth->bst",
            query_embeddings,
            document_embeddings,
        )
        # sum(1) will sum up in sequence_length dim, later divide the sum of non-pad token
        late_interactions = late_interactions.max(2).values.sum(1)
        if query_attention_mask is not None:
            late_interactions = late_interactions / query_attention_mask[:, 1:].sum(-1, keepdim=False)
        else:  # Or length of token sequence
            late_interactions = late_interactions / query_embeddings.size(1)
        return late_interactions

    def save_pretrained(self, save_directory: Union[str, os.PathLike], safe_serialization: bool = True):
        logger.info("Save model to {}".format(save_directory))
        state_dict_fn = lambda state_dict: type(state_dict)({k: v.clone().cpu() for k, v in state_dict.items()})
        state_dict = self.model.state_dict()
        self.model.save_pretrained(
            save_directory, state_dict=state_dict_fn(state_dict), safe_serialization=safe_serialization
        )
        torch.save(state_dict_fn(self.linear.state_dict()), os.path.join(save_directory, 'colbert_linear.pt'))
        self.tokenizer.save_pretrained(save_directory)

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_path: str,
        colbert_dim: int = 1024,
        pretrained_colbert_linear_name: str = 'colbert_linear.pt',
        loss_fn: Union[nn.Module, Callable] = ColbertLoss(),
        trust_remote_code: bool = True,
        use_fp16: bool = False,
        device: Optional[str] = None,
        **kwargs,
    ):
        if not os.path.exists(model_name_or_path):
            """To load linear layer weight, manually download the model so the model_name_or_path will always be path"""
            from huggingface_hub import snapshot_download

            cache_dir = os.getenv('HF_HUB_CACHE')
            model_name_or_path = snapshot_download(
                repo_id=model_name_or_path,
                cache_dir=cache_dir,
                ignore_patterns=['flax_model.msgpack', 'rust_model.ot', 'tf_model.h5'],
            )

        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, trust_remote_code=trust_remote_code)
        model = AutoModel.from_pretrained(model_name_or_path, trust_remote_code=trust_remote_code, **kwargs)

        linear_layer = nn.Linear(model.config.hidden_size, colbert_dim)
        if os.path.exists(path=os.path.join(model_name_or_path, pretrained_colbert_linear_name)):
            logger.info(
                f'Loading colbert_linear pretrained weight from {model_name_or_path}, colbert_dim={colbert_dim}'
            )
            colbert_state_dict = torch.load(
                os.path.join(model_name_or_path, pretrained_colbert_linear_name), map_location='cpu'
            )
            linear_layer.load_state_dict(colbert_state_dict)
        else:
            logger.info(f'Xavier uniform random colbert linear layer,  colbert_dim={colbert_dim}')
            torch.nn.init.xavier_uniform_(tensor=linear_layer.weight)

        if os.path.exists(path=os.path.join(model_name_or_path, "metadata.json")):
            with open(file=os.path.join(model_name_or_path, "metadata.json"), mode="r") as f:
                metadata = json.load(fp=f)
            max_length = metadata["max_length"]
            print(max_length)

        if use_fp16:
            logger.info('Set model to fp16, please note that if you want fp16 during training, set training_args fp16')
            model.half()
            linear_layer.half()

        ranker = cls(
            model=model,
            tokenizer=tokenizer,
            linear_layer=linear_layer,
            device=device,
            loss_fn=loss_fn,
        )
        return ranker


class LLMRanker(AutoModelForRanking):
    def __init__(self, task_prompt: Optional[str] = None, token='Yes', **kwargs):
        super(LLMRanker, self).__init__(**kwargs)
        if task_prompt is None:
            task_prompt = (
                "Given a query A and a passage B, determine whether the passage contains an answer to the query"
                "by providing a prediction of either 'Yes' or 'No'."
            )
        self.task_prompt = task_prompt
        self.prompt_inputs = self.tokenizer(self.task_prompt, return_tensors=None, add_special_tokens=False)[
            'input_ids'
        ]
        sep = "\n"
        self.sep_inputs = self.tokenizer(sep, return_tensors=None, add_special_tokens=False)['input_ids']
        self.token_loc = self.tokenizer(token, add_special_tokens=False)['input_ids'][0]

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        labels: Optional[torch.Tensor] = None,
        return_dict: Optional[bool] = True,
        **kwargs,
    ):
        model_output = self.model(input_ids, attention_mask, output_hidden_states=True)
        loss = self.loss_fn(model_output.logits, labels)
        outputs_dict = dict()
        outputs_dict['logits'] = model_output.logits
        outputs_dict['loss'] = loss
        return outputs_dict

    def preprocess_pair(self, batch_sentence_pair: List[List[str]], max_length: int, **kwargs):
        collator = LLMRerankCollator(tokenizer=self.tokenizer, prompt=self.task_prompt, max_length=max_length)
        batch_inputs = collator(batch_sentence_pair)

        batch_inputs = {
            'input_ids': batch_inputs['input_ids'].to(self.device),
            'attention_mask': batch_inputs['attention_mask'].to(self.device),
        }
        return batch_inputs

    @torch.no_grad()
    def compute_score(
        self,
        sentence_pairs: Union[List[Tuple[str]], Tuple[str]],
        batch_size: int = 16,
        max_length: int = 256,
        normalize: bool = False,
        show_progress_bar: bool = None,
        **kwargs,
    ):
        self.model.eval()

        if isinstance(sentence_pairs[0], str):
            sentence_pairs = [sentence_pairs]

        length_sorted_idx = np.argsort([-self._text_length(q) - self._text_length(p) for q, p in sentence_pairs])
        sentences_sorted = [sentence_pairs[idx] for idx in length_sorted_idx]

        if self.query_instruction or self.document_instruction:
            sentences_sorted = [
                (self.query_instruction + pair[0], self.document_instruction + pair[1]) for pair in sentences_sorted
            ]

        all_scores: List[float] = []
        for batch_start in tqdm(
            range(0, len(sentences_sorted), batch_size), desc='Scoring', disable=not show_progress_bar
        ):
            batch_sentences = sentences_sorted[batch_start : batch_start + batch_size]
            batch_on_device = self.preprocess_pair(batch_sentences, max_length=max_length)
            outputs = self.model(**batch_on_device, output_hidden_states=True)
            scores = self.score(outputs.logits, batch_on_device['attention_mask'])

            if normalize:
                scores = torch.sigmoid(scores)
            all_scores.extend(scores.cpu().float().tolist())

        all_scores = [all_scores[idx] for idx in np.argsort(length_sorted_idx)]

        if len(all_scores) == 1:
            return all_scores[0]

        return all_scores

    def score(self, logits, attention_mask: torch.Tensor):
        scores = AutoPooling('last')(logits, attention_mask)
        scores = scores[:, self.token_loc]
        return scores


class DocumentSplitter(object):
    """
    Rerank the long document
    - https://github.com/netease-youdao/BCEmbedding/blob/master/BCEmbedding/models/utils.py
    """

    def __init__(self, chunk_size: int, chunk_overlap: int = 0, max_chunks_per_doc: int = 32):
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.max_chunks_per_doc = max_chunks_per_doc

    def create_documents(self, query, documents, tokenizer):
        res_merge_inputs = []
        res_merge_inputs_pids = []

        query_inputs = tokenizer.encode_plus(query, truncation=False, padding=False)
        sep_id = tokenizer.sep_token_id
        doc_max_length = self.chunk_size - len(query_inputs['input_ids']) - 2

        for pid, document in enumerate(documents):
            document_inputs = tokenizer.encode_plus(document, truncation=False, padding=False, add_special_tokens=False)
            doc_inputs_length = len(document_inputs['input_ids'])

            if doc_inputs_length <= doc_max_length:
                qc_merge_inputs = self._merge_inputs(query_inputs, document_inputs, sep_id)
                res_merge_inputs.append(qc_merge_inputs)
                res_merge_inputs_pids.append(pid)
            else:
                start_id = 0
                while start_id < doc_inputs_length:
                    end_id = start_id + doc_max_length
                    sub_document_inputs = {k: v[start_id:end_id] for k, v in document_inputs.items()}
                    start_id = end_id - self.chunk_overlap if end_id < doc_inputs_length else end_id

                    qp_merge_inputs = self._merge_inputs(query_inputs, sub_document_inputs, sep_id)
                    res_merge_inputs.append(qp_merge_inputs)
                    res_merge_inputs_pids.append(pid)
        return res_merge_inputs, res_merge_inputs_pids

    def _merge_inputs(self, chunk1_raw, chunk2, sep_id: int):
        chunk1 = deepcopy(chunk1_raw)

        chunk1['input_ids'].append(sep_id)
        chunk1['input_ids'].extend(chunk2['input_ids'])
        chunk1['input_ids'].append(sep_id)

        chunk1['attention_mask'].append(chunk2['attention_mask'][0])
        chunk1['attention_mask'].extend(chunk2['attention_mask'])
        chunk1['attention_mask'].append(chunk2['attention_mask'][0])

        if 'token_type_ids' in chunk1:
            token_type_ids = [1 for _ in range(len(chunk2['token_type_ids']) + 2)]
            chunk1['token_type_ids'].extend(token_type_ids)
        return chunk1



import torch
import torch.nn as nn
import torch.nn.functional as F


class AutoPooling(nn.Module):
    """Pooling class to transfer input to 1D embedding"""

    def __init__(self, pooling_method: str, **kwargs) -> None:
        super().__init__()
        if pooling_method in ["mean", 'avg', 'average']:
            self.pooling = MeanPooling()
        elif pooling_method in ["cls", 'first']:
            self.pooling = ClsTokenPooling()
        elif pooling_method == "weighted":
            self.pooling = WeightedLayerPooling()
        elif pooling_method in ['eos', "last"]:
            self.pooling = LastTokenPooling()
        else:
            raise ValueError(
                f"pooling_method {pooling_method} is not a valid method: 'cls', 'mean', 'weighted', 'last'"
            )

    def forward(self, last_hidden_state: torch.Tensor, attention_mask: torch.Tensor) -> torch.Tensor:
        return self.pooling(last_hidden_state, attention_mask)


class MeanPooling(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, last_hidden_state: torch.Tensor, attention_mask: torch.Tensor) -> torch.Tensor:
        """
        last hidden state: First element of model_output contains all token embeddings
        """
        # last_hidden_state = model_output[0]
        attention_mask = attention_mask.unsqueeze(-1).float()
        sum_embeddings = torch.sum(last_hidden_state * attention_mask, 1)
        sum_mask = torch.clamp(attention_mask.sum(1), min=1e-9)
        mean_embeddings = sum_embeddings / sum_mask
        return mean_embeddings


class ClsTokenPooling(nn.Module):
    """CLS Pooling"""

    def __init__(self):
        super().__init__()

    def forward(
        self,
        last_hidden_state: torch.Tensor,
        attention_mask: torch.Tensor,
    ) -> torch.Tensor:
        # Take embeddings of first token per sample
        # left_padding = (attention_mask[:, -1].sum() == attention_mask.shape[0])
        # if left_padding:
        #     return last_hidden_states[:, -1]

        cls_token = last_hidden_state[:, 0]
        return cls_token


class LastTokenPooling(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(
        self,
        last_hidden_state: torch.Tensor,
        attention_mask: torch.Tensor,
    ) -> torch.Tensor:
        left_padding = attention_mask[:, -1].sum() == attention_mask.shape[0]
        if left_padding:
            emb = last_hidden_state[:, -1]
        else:
            sequence_lengths = attention_mask.sum(dim=1) - 1
            batch_size = last_hidden_state.shape[0]
            emb = last_hidden_state[
                torch.arange(batch_size, device=last_hidden_state.device),
                sequence_lengths,
            ]
        return emb


class AttentionPooling(nn.Module):
    def __init__(self):
        super(AttentionPooling, self).__init__()

    def forward(self, last_hidden_state: torch.Tensor, attention_mask: torch.Tensor) -> torch.Tensor:
        return last_hidden_state[:, 0, :]


class WeightedLayerPooling(nn.Module):
    def __init__(self, num_hidden_layers: int, layer_start: int = 4, layer_weights=None):
        super().__init__()
        self.layer_start = layer_start
        self.num_hidden_layers = num_hidden_layers
        self.layer_weights = (
            layer_weights
            if layer_weights is not None
            else nn.Parameter(torch.tensor([1] * (num_hidden_layers + 1 - layer_start), dtype=torch.float))
        )

    def forward(self, all_hidden_states: torch.Tensor) -> torch.Tensor:
        all_layer_embedding = all_hidden_states[self.layer_start :, :, :, :]
        weight_factor = self.layer_weights.unsqueeze(-1).unsqueeze(-1).unsqueeze(-1).expand(all_layer_embedding.size())
        weighted_average = (weight_factor * all_layer_embedding).sum(dim=0) / self.layer_weights.sum()
        return weighted_average


class GeMText(nn.Module):
    """GeM Pooling for NLP
    Generalized mean: https://arxiv.org/abs/1711.02512
    """

    def __init__(self, dim, p: int = 3, eps: float = 1e-6):
        super(GeMText, self).__init__()
        self.dim = dim
        self.p = nn.Parameter(torch.ones(1) * p)
        self.eps = eps

    def forward(
        self,
        x: torch.Tensor,
        attention_mask: torch.Tensor,
    ) -> torch.Tensor:
        attention_mask_expanded = attention_mask.unsqueeze(-1).expand(x.shape)
        x = (x.clamp(min=self.eps) * attention_mask_expanded).pow(self.p).sum(self.dim)
        ret = x / attention_mask_expanded.sum(self.dim).clip(min=self.eps)
        ret = ret.pow(1 / self.p)
        return ret


def gem(x, p=3, eps=1e-6):
    return F.avg_pool2d(x.clamp(min=eps).pow(p), (x.size(-2), x.size(-1))).pow(1.0 / p)


class GeM(nn.Module):
    def __init__(self, p=3, eps: float = 1e-6, p_trainable: bool = True):
        super(GeM, self).__init__()
        if p_trainable:
            self.p = nn.Parameter(torch.ones(1) * p)
        else:
            self.p = p
        self.eps = eps

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return gem(x, p=self.p, eps=self.eps)

    def __repr__(self):
        return (
            self.__class__.__name__
            + "("
            + "p="
            + "{:.4f}".format(self.p.data.tolist()[0])
            + ", "
            + "eps="
            + str(self.eps)
            + ")"
        )


class AdaptiveGeM(nn.Module):
    def __init__(self, size=(1, 1), p: int = 3, eps: float = 1e-6):
        super().__init__()
        self.size = size
        self.p = nn.Parameter(torch.ones(1) * p)
        self.eps = eps

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.adaptive_avg_pool2d(x.clamp(min=self.eps).pow(self.p), self.size).pow(1.0 / self.p)

    def __repr__(self):
        return f"AdaptiveGeM(size={self.size}, p={self.p}, eps={self.eps})"


class TopKPooling(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self):
        return


class SumPooling(nn.Module):
    def forward(self, x, x_mask=None):
        if x_mask is None or x_mask.data.sum() == 0:
            return torch.sum(x, 1)
        else:
            x_mask = x_mask.unsqueeze(-1).expand(x.size())
            x.data.masked_fill_(x_mask.data, 0.0)
            return torch.sum(x, 1)


class FMPooling(nn.Module):
    def __init__(self):
        super(FMPooling, self).__init__()
        self.sum_pooling = SumPooling()

    def forward(self, x, x_mask=None):
        summed_emb = self.sum_pooling(x, x_mask)
        summed_emb_square = summed_emb**2

        squared_emb = x**2
        squared_sum_emb = self.sum_pooling(squared_emb, x_mask)

        y_second_order = 0.5 * (summed_emb_square - squared_sum_emb)
        return y_second_order



"""Embedding fine tune pipeline"""

import logging
import os
import pickle
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional

import torch
from torch.utils.data import DataLoader
from transformers import AutoTokenizer, HfArgumentParser, TrainingArguments, set_seed

from ..data import (
    EncodeCollator,
    EncodeDataset,
    PairCollator,
    RetrievalTrainDataset,
    TripletCollator,
)
from ..losses import AutoLoss, InfoNCE, SimCSE, TripletLoss
from ..models.embedding_auto import AutoModelForEmbedding
from ..trainer import RetrievalTrainer

# os.environ["WANDB_LOG_MODEL"] = "false"
logger = logging.getLogger(__name__)


@dataclass
class ModelArguments:
    model_name_or_path: str = field(
        metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models"}
    )
    config_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained config name or path if not the same as model_name"}
    )
    tokenizer_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained tokenizer name or path if not the same as model_name"}
    )
    cache_dir: Optional[str] = field(
        default=None, metadata={"help": "Where do you want to store the pretrained models downloaded from s3"}
    )
    causal_lm: bool = field(default=False, metadata={'help': "Whether the model is a causal lm or not"})
    lora_path: Optional[str] = field(default=None, metadata={'help': "Lora adapter save path"})


@dataclass
class DataArguments:
    data_name_or_path: str = field(default=None, metadata={"help": "Path to train data"})
    train_group_size: int = field(default=2)
    unfold_each_positive: bool = field(default=False)
    query_max_length: int = field(
        default=32,
        metadata={
            "help": "The maximum total input sequence length after tokenization for passage. Sequences longer "
            "than this will be truncated, sequences shorter will be padded."
        },
    )
    document_max_length: int = field(
        default=128,
        metadata={
            "help": "The maximum total input sequence length after tokenization for passage. Sequences longer "
            "than this will be truncated, sequences shorter will be padded."
        },
    )
    query_instruction: str = field(default=None, metadata={"help": "instruction for query"})
    document_instruction: str = field(default=None, metadata={"help": "instruction for document"})
    query_key: str = field(default=None)
    positive_key: str = field(default='positive')
    negative_key: str = field(default='negative')
    is_query: bool = field(default=False)
    encoding_save_file: str = field(default='embed.pkl')

    def __post_init__(self):
        # self.data_name_or_path = 'json'
        self.dataset_split = 'train'
        self.dataset_language = 'default'

        if self.data_name_or_path is not None:
            if not os.path.isfile(self.data_name_or_path) and not os.path.isdir(self.data_name_or_path):
                info = self.data_name_or_path.split('/')
                self.dataset_split = info[-1] if len(info) == 3 else 'train'
                self.data_name_or_path = "/".join(info[:-1]) if len(info) == 3 else '/'.join(info)
                self.dataset_language = 'default'
                if ':' in self.data_name_or_path:
                    self.data_name_or_path, self.dataset_language = self.data_name_or_path.split(':')


@dataclass
class RetrieverTrainingArguments(TrainingArguments):
    train_type: str = field(default='pairwise', metadata={'help': "train type of point, pair, or list"})
    negatives_cross_device: bool = field(default=False, metadata={"help": "share negatives across devices"})
    temperature: Optional[float] = field(default=0.02)
    fix_position_embedding: bool = field(
        default=False, metadata={"help": "Freeze the parameters of position embeddings"}
    )
    pooling_method: str = field(default='cls', metadata={"help": "the pooling method, should be cls or mean"})
    normalized: bool = field(default=True)
    loss_fn: str = field(default='infonce')
    use_inbatch_negative: bool = field(default=True, metadata={"help": "use documents in the same batch as negatives"})
    remove_unused_columns: bool = field(default=False)
    use_lora: bool = field(default=False)
    use_bnb_config: bool = field(default=False)
    do_encode: bool = field(default=False, metadata={"help": "run the encoding loop"})
    report_to: Optional[List[str]] = field(
        default="none", metadata={"help": "The list of integrations to report the results and logs to."}
    )


def main():
    parser = HfArgumentParser((ModelArguments, DataArguments, RetrieverTrainingArguments))
    model_args, data_args, training_args = parser.parse_args_into_dataclasses()
    model_args: ModelArguments
    data_args: DataArguments
    training_args: TrainingArguments

    if (
        os.path.exists(training_args.output_dir)
        and os.listdir(training_args.output_dir)
        and training_args.do_train
        and not training_args.overwrite_output_dir
    ):
        raise ValueError(
            f"Output directory ({training_args.output_dir}) already exists and is not empty. "
            "Use --overwrite_output_dir to overcome."
        )

    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO if training_args.local_rank in [-1, 0] else logging.WARN,
    )
    logger.warning(
        "Process rank: %s, device: %s, n_gpu: %s, distributed training: %s, 16-bits training: %s",
        training_args.local_rank,
        training_args.device,
        training_args.n_gpu,
        bool(training_args.local_rank != -1),
        training_args.fp16,
    )
    logger.info("Training/evaluation parameters %s", training_args)
    logger.info("Model parameters %s", model_args)
    logger.info("Data parameters %s", data_args)

    set_seed(training_args.seed)

    tokenizer = AutoTokenizer.from_pretrained(
        model_args.tokenizer_name if model_args.tokenizer_name else model_args.model_name_or_path,
        cache_dir=model_args.cache_dir,
        use_fast=False,
    )
    if training_args.use_bnb_config:
        from transformers import BitsAndBytesConfig

        logger.info('Use quantization bnb config')
        quantization_config = BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_use_double_quant=True,
            bnb_4bit_quant_type="nf4",
            bnb_4bit_compute_dtype=torch.bfloat16,
        )
    else:
        quantization_config = None

    if training_args.do_train:
        model = AutoModelForEmbedding.from_pretrained(
            model_name_or_path=model_args.model_name_or_path,
            pooling_method=training_args.pooling_method,
            use_lora=training_args.use_lora,
            quantization_config=quantization_config,
        )

        loss_fn = AutoLoss(
            loss_name=training_args.loss_fn,
            loss_kwargs={
                'use_inbatch_negative': training_args.use_inbatch_negative,
                'temperature': training_args.temperature,
            },
        )

        model = model.set_train_type(
            "pairwise",
            loss_fn=loss_fn,
        )

        train_dataset = RetrievalTrainDataset(
            args=data_args,
            tokenizer=tokenizer,
            positive_key=data_args.positive_key,
            negative_key=data_args.negative_key,
        )
        logger.info(f"Total training examples: {len(train_dataset)}")

        trainer = RetrievalTrainer(
            model=model,
            args=training_args,
            train_dataset=train_dataset,
            data_collator=TripletCollator(
                tokenizer,
                query_max_length=data_args.query_max_length,
                document_max_length=data_args.document_max_length,
                positive_key=data_args.positive_key,
                negative_key=data_args.negative_key,
            ),
        )

        Path(training_args.output_dir).mkdir(parents=True, exist_ok=True)

        trainer.train()
        # trainer.save_model(training_args.output_dir)
        model.save_pretrained(training_args.output_dir)

        if trainer.is_world_process_zero():
            tokenizer.save_pretrained(training_args.output_dir)

    if training_args.do_encode:
        model = AutoModelForEmbedding.from_pretrained(
            model_name_or_path=model_args.model_name_or_path,
            pooling_method=training_args.pooling_method,
            use_lora=training_args.use_lora,
            quantization_config=quantization_config,
            lora_path=model_args.lora_path,
        )

        max_length = data_args.query_max_length if data_args.is_query else data_args.document_max_length
        logger.info(f'Encoding will be saved in {training_args.output_dir}')

        encode_dataset = EncodeDataset(args=data_args, tokenizer=tokenizer, max_length=max_length, text_key='text')
        logger.info(f"Number of train samples: {len(encode_dataset)}, max_length: {max_length}")

        encode_loader = DataLoader(
            encode_dataset,
            batch_size=training_args.per_device_eval_batch_size,
            collate_fn=EncodeCollator(tokenizer, max_length=max_length, padding='max_length'),
            shuffle=False,
            drop_last=False,
            num_workers=training_args.dataloader_num_workers,
        )

        embeddings = model.encode(encode_loader, show_progress_bar=True, convert_to_numpy=True)
        lookup_indices = list(range(len(encode_dataset)))

        with open(os.path.join(training_args.output_dir, data_args.encoding_save_file), 'wb') as f:
            pickle.dump((embeddings, lookup_indices), f)


if __name__ == "__main__":
    main()



"""
`FaceNet: A Unified Embedding for Face Recognition and Clustering
<https://arxiv.org/abs/1503.03832>`_
"""

from typing import Optional

import torch
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F


class TripletLoss(nn.Module):
    """
    https://omoindrot.github.io/triplet-loss
    """

    def __init__(
        self,
        temperature: float = 0.05,
        margin: float = 0.0,
        negatives_cross_device: bool = False,
        batch_hard: bool = False,
        **kwargs
    ):
        super().__init__()
        self.temperature = temperature
        self.margin = margin
        self.negatives_cross_device = negatives_cross_device
        self.batch_hard = batch_hard
        if self.negatives_cross_device:
            if not dist.is_initialized():
                raise ValueError("Cannot do negatives_cross_device without distributed training")
            self.rank = dist.get_rank()
            self.world_size = dist.get_world_size()

    def forward(
        self,
        query_embeddings: torch.Tensor,
        pos_embeddings: torch.Tensor,
        neg_embeddings: torch.Tensor,
        margin: Optional[float] = None,
    ):
        if margin is not None:
            # dynamic margin
            self.set_margin(margin=margin)

        if self.negatives_cross_device:
            pos_embeddings = self._dist_gather_tensor(pos_embeddings)
            neg_embeddings = self._dist_gather_tensor(neg_embeddings)

        pos_similarity = torch.cosine_similarity(query_embeddings, pos_embeddings, dim=-1)
        pos_similarity = pos_similarity / self.temperature
        neg_similarity = torch.cosine_similarity(
            query_embeddings.unsqueeze(1),
            neg_embeddings.unsqueeze(0),
            dim=-1,
        )
        neg_similarity = neg_similarity / self.temperature
        similarity_diff = pos_similarity.unsqueeze(1) - neg_similarity
        loss = -torch.log(torch.sigmoid(similarity_diff)).mean()
        return loss

    def set_margin(self, margin: float):
        self.margin = margin

    def _dist_gather_tensor(self, t: Optional[torch.Tensor]):
        if t is None:
            return None
        t = t.contiguous()

        all_tensors = [torch.empty_like(t) for _ in range(self.world_size)]
        dist.all_gather(all_tensors, t)

        all_tensors[self.rank] = t
        all_tensors = torch.cat(all_tensors, dim=0)

        return all_tensors


class TripletCosineSimilarity(nn.Module):
    def __init__(self, temperature: float = 1.0, margin: float = 0.50):
        super().__init__()
        self.temperature = temperature
        self.margin = margin
        self.distance_metric = lambda x, y: F.pairwise_distance(x, y, p=2)

    def forward(self, query_embedding: torch.Tensor, pos_embedding: torch.Tensor, neg_embedding: torch.Tensor):
        distance_pos = self.distance_metric(query_embedding, pos_embedding)
        distance_neg = self.distance_metric(query_embedding, neg_embedding)

        losses = F.relu(distance_pos - distance_neg + self.margin)
        return losses.mean()
